-- DONE SYNTAX CHANGE
-----------------------------------------------------------------------------
-- |
-- Module      :  LRSX.Unification.UnificationProblem
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--  
-- The data type for unification problems which is used for the unification algorithm
--
-----------------------------------------------------------------------------
-- ====================================================
module LRSX.Unification.UnificationProblem where
import Data.List
import Data.Maybe
import LRSX.Language.Syntax
import LRSX.Language.Fresh
import LRSX.Language.ContextDefinition


-- | the function used for checking the NCCs, 
--   abstracted here to change it into other functions
--   at the moment it is synonymous to 'capturedInHole'
checkBVorNCC :: Context -> [VarLift]
checkBVorNCC e = capturedInHole e

 
-- ============================================================================
-- Data-type for a unification problem
-- ============================================================================
 
data Problem = Problem {
                       solution      :: Solution,               -- ^ the solution, generated by unification
                       gamma         :: Gamma,                  -- ^ the unification equations
                       delta         :: Delta,                  -- ^ the constraints
                       nabla :: Delta,                  -- ^ fixed (given) constraints, only used for matching, not for unification
                       fresh         :: [String],               -- ^ a list of fresh names
                       functions     :: FunInfo,                -- ^ the function symbols together with their arity and a list of strict positions
                       inputGamma    :: Gamma,                  -- ^ the input equations
                       inputDelta    :: Delta,                  -- ^ the input constraints in delta
                       alphaBase     :: [(SyntaxComponent,SubstItem)],   -- ^ freshness constraints on an alpha renaming, not used by the unification, but for matching
                       mpLogging     :: String,                          -- ^ a String to store logging information 
                       expandDelta4Constraints :: Bool,                  -- ^ True, if E1;E2 = EE[z,s] should be expanded, if there is an NCC (E1,E2) or (E2,E1)
                       expandAllContexts :: Bool,                        -- ^ True, if all (and not only needed) context variables are guessed empty/non-empty
                                                                         --   set to True, this may help when joining overlaps with matching
                       expandAllEnvs :: Bool,
                       contextDefinitions :: CtxtDef
                  } 
 deriving(Show)                    

-- compute the strict positions of a function symbol
-- getStrictPos :: Problem -> Name -> Maybe [Int]
-- getStrictPos pb f = 
--  case [pos | (n,_,pos,_) <- (functions pb), n == f] of
--  []      -> Nothing
--  ([pos]) -> Just pos
--  (_:_)   -> error "in getStrictPos: Positions are not unique?"

-- ============================================================================
-- Gamma: multisets of different equations:
-- ============================================================================
data Gamma = Gamma {varLiftEq     :: [Equation VarLift]     -- ^ variable equations
                   ,expressionEq  :: [Equation Expression]  -- ^ expression equations, note that it is forbidden to equate contexts with expressions, 
                                                            --   which is not ensured by the modelling                      
                   ,environmentEq :: [Equation Environment] -- ^ enviroment equations
                   ,bindingEq     :: [Equation Binding]     -- ^ binding equations
                   }
                   
-- | check whether an environment is empty                   
nullGamma :: Gamma -> Bool
nullGamma (Gamma{varLiftEq = d1, expressionEq =d2, environmentEq=d3, bindingEq = d4}) = 
   and [null d1,null d2,null d3,null d4]                   
   
-- | construct an empty enviroment   
emptyGamma :: Gamma   
emptyGamma = (Gamma{varLiftEq = [], expressionEq =[], environmentEq=[], bindingEq = []})

-- Show-instance                   
instance Show Gamma where
 show sol   = "{" ++ (concat $ intersperse "," (filter (not.null) $  [ showMap (varLiftEq sol), showMap (expressionEq sol), showMap (environmentEq sol), showMap (bindingEq sol)])) ++ "}"
  where
   showMap [] = ""
   showMap mp =   concat $ intersperse "," $   [show eq | eq <-   mp] 

-- | compute all variables occuring in gamma (meta and concrete variables)
allVariablesInGamma :: Gamma -> [VarLift]
allVariablesInGamma gam = concat $
   [ [l,r] | (l :=?=: r) <- varLiftEq gam]
   ++
   [ (allVariables l) ++ (allVariables r) | (l :=?=: r) <- expressionEq gam]
   ++
   [ (xl):(xr):((allVariables l) ++ (allVariables r))  | ((xl:=l) :=?=: (xr:=r)) <- bindingEq gam]
   ++
   [ allVariablesInEnv env1 ++ allVariablesInEnv env2 | (env1 :=?=: env2) <- environmentEq gam]

instance Constantify a => Constantify (Equation a) where
  constantify (a :=?=: b) = (constantify a) :=?=: (constantify b)
  deconstantify (a :=?=: b) = (deconstantify a) :=?=: (deconstantify b)
  
instance Constantify Gamma where
  constantify gamma = gamma{varLiftEq = map constantify (varLiftEq gamma)
                          
                           ,expressionEq = map constantify (expressionEq gamma)
                           ,bindingEq  = map constantify (bindingEq gamma)
                           ,environmentEq = map constantify (environmentEq gamma)
  } 
  deconstantify gamma = gamma{varLiftEq = map deconstantify (varLiftEq gamma)
                          
                           ,expressionEq = map deconstantify (expressionEq gamma)
                           ,bindingEq  = map deconstantify (bindingEq gamma)
                           ,environmentEq = map deconstantify (environmentEq gamma)
  } 

showNonCapConstraint (Left e,e') = "(" ++ show e ++ "," ++ show e' ++ ")"
showNonCapConstraint (Right e,e') = "[" ++ show e ++ "," ++ show e' ++ "]"
showNCCs :: NonCaptureConstraints -> String
showNCCs   = intercalate "," . map showNonCapConstraint

instance Constantify Delta where
  constantify delta = delta{delta1 = constantify (delta1 delta)
                           ,delta2 = delta2 delta
                           ,delta3 = [(constantify l,constantify r) | (l,r) <-delta3 delta]
                           ,delta4 = constantify (delta4 delta)
  }
  deconstantify delta = delta{delta1 = deconstantify (delta1 delta)
                           ,delta2 = delta2 delta
                           ,delta3 = [(deconstantify l,deconstantify r) | (l,r) <-delta3 delta]
                           ,delta4 = deconstantify (delta4 delta)
  }
-- Substitute-instance
instance Substitute Gamma where
 subst e1 e2  gamma = substG e1 e2 gamma 
  where 
    substG e1 e2 gamma =
      gamma {varLiftEq = subst e1 e2 (varLiftEq gamma)
        
            ,expressionEq = subst e1 e2 (expressionEq gamma)
            ,environmentEq = subst e1 e2 (environmentEq gamma)
            ,bindingEq = subst e1 e2 (bindingEq gamma)
            }
 substEmpty n gamma = substG n gamma 
  where 
    substG n gamma =
      gamma {varLiftEq = substEmpty n (varLiftEq gamma)
      
            ,expressionEq = substEmpty n (expressionEq gamma)
            ,environmentEq = substEmpty n (environmentEq gamma)
            ,bindingEq     = substEmpty n (bindingEq gamma)
            }
 substEnv env n gamma = 
      gamma {varLiftEq = substEnv env n (varLiftEq gamma)
            
            ,expressionEq = substEnv env n (expressionEq gamma)
            ,environmentEq = substEnv env n (environmentEq gamma)
            ,bindingEq = substEnv env n (bindingEq gamma)
            }
 substChainVar env n gamma = 
      gamma {varLiftEq = substChainVar env n (varLiftEq gamma)
            
            ,expressionEq = substChainVar env n (expressionEq gamma)
            ,environmentEq = substChainVar env n (environmentEq gamma)
            ,bindingEq = substChainVar env n (bindingEq gamma)
            }
 substEnvConstraint env n gamma = 
      gamma {varLiftEq = substEnvConstraint env n (varLiftEq gamma)
          
            ,expressionEq = substEnvConstraint env n (expressionEq gamma)
            ,environmentEq = substEnvConstraint env n (environmentEq gamma)
            ,bindingEq = substEnvConstraint env n (bindingEq gamma)
            }
   
-- solved checks whether all equations are solved, i.e. all sets are empty
solved :: Gamma -> Bool
solved gamma = 
 and [null $ varLiftEq gamma,  null $ expressionEq gamma, null $ environmentEq gamma, null $ bindingEq gamma] 

 
 
-- Equations:
data Equation a = a :=?=: a
 deriving(Eq)
-- Show-instance
instance Show a => Show (Equation a) where
 show (a :=?=: b) = show a ++ " =?= " ++ show b
instance HasMetaVars a => HasMetaVars (Equation a) where
 computeMetaVars (a :=?=: b) = combineMetaVarsList [computeMetaVars a, computeMetaVars b]
-- Substitute-instance
instance Substitute a => Substitute (Equation a) where
  subst e1 e2 (a :=?=: b) = (subst e1 e2 a) :=?=: (subst e1 e2 b)
  substEmpty n (a :=?=: b) = (substEmpty n  a) :=?=: (substEmpty n b)
  substEnv env n ( a :=?=: b) = (substEnv env n a) :=?=: (substEnv env n b)
  substChainVar env v ( a :=?=: b) = (substChainVar env v a) :=?=: (substChainVar env v b)
  substEnvConstraint env v (a :=?=: b)  = (substEnvConstraint env v a) :=?=: (substEnvConstraint env v b)
instance HasFnSymbol a => HasFnSymbol (Equation a) where
 getFnSymbols (a :=?=: b) = nubFnSyms $ getFnSymbols a ++ getFnSymbols b
 
commute :: Equation a -> Equation a 
commute (left :=?=: right) = right :=?=: left  
 
  
type EquationsWithConstraints = ([Equation Expression],[Expression],[EnvVar],NonCaptureConstraints)
type EquationsWithTwoConstraints = ([Equation Expression],[Expression],[EnvVar],NonCaptureConstraints,[Expression],[EnvVar],NonCaptureConstraints)
-- List of equations, list of nonempty contexts, list of nonempty environments
  
-- check constrained equations
-- check :: EquationsWithConstraints -> Maybe String -- Nothing = ok, Just err = Fail
check pb = 
 let eqs = expressionEq (gamma pb)
     ctxtcon = delta1 (delta pb)
     nonemptyenvvars = delta2 (delta pb)
     capcon = delta3 (delta pb) 
 in checkGamma (Gamma{expressionEq=eqs,bindingEq=[],environmentEq=[],varLiftEq=[]},ctxtcon,nonemptyenvvars,capcon)

  
checkGammaUnifInp (gamma,ctxtcon,nonemptyenvvars)
     | ((v,a,b):_) <- checkDoubleVars = Just $ a ++ " " ++ v ++ " occurs also as " ++ b 
     | ((u,v,eq):_) <- checkChains = Just $ "chain-variables on both sides of the equation " ++ eq ++ " (" ++ u ++ " on the left hand side and " ++ v ++ " on the right hand side)"
     | ((v:_,dec):_) <- checkOtherVars = Just $ dec ++ " " ++ v ++ " occurs more than once in the unification problem"
     | Just err <- checkLVCGamma gamma = Just $ "LVC not valid:" ++ err
     | otherwise = Nothing
   where  
    nonemptycxtvars = [v | MetaCtxt s srt v <- ctxtcon]
    eq1 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- varLiftEq gamma]
    eq3 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- expressionEq gamma]
    eq4 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- bindingEq gamma]
    eq5 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- environmentEq gamma]
    allVars = combineMetaVarsList $
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq1]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq3]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq4]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq5]
    checkNonEmptyCtxtVars = [v | v <- nonemptycxtvars, v `notElem` (ctxtvars allVars)]
    checkNonEmptyEnvVars = [v | Env s v <- nonemptyenvvars, v `notElem` (envvars allVars)]
    checkDoubleVars =    [(v,"context-variable","variable")      | v <- (ctxtvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"context-variable","expression variable")        | v <- (ctxtvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"context-variable","environment-variable")       | v <- (ctxtvars allVars), v `elem` (envvars allVars)]
                      ++ [(v,"context-variable","chain-variable")             | v <- (ctxtvars allVars), v `elem` (chainvars allVars)]
                      ++ [(v,"chain-variable","bound variable")        | v <- (chainvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"chain-variable","expression variable")          | v <- (chainvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"chain-variable","environment-variable")         | v <- (chainvars allVars), v `elem` (envvars allVars)]
                      ++ [(v,"environment-variable","variable")  | v <- (envvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"environment-variable","expression variable")    | v <- (envvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"expression-variable","bound variable")   | v <- (exprvars allVars), v `elem` (liftvars allVars)]
    checkEquations = concat $
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq1, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq3, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"let-bound variable",eq)    | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq4, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"let-bound variable",eq)    | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq5, let eq = show (l :=?=: r)]                   
                         
    checkChains =     [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq1, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq3, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq4, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq5, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                                      
    checkExprVars = filter (\x -> length x > 2) $ group $ sort $ exprvars allVars
    checkOtherVars = concat [zip (filter (\x -> length x > 1) $ group $ sort $ vars) (repeat dec)
                         | (vars,dec) <- [(envvars allVars,"environment-variable"),(ctxtvars allVars,"context-variable"),(chainvars allVars,"chain-variable")]]

checkGamma (gamma,ctxtcon,nonemptyenvvars,capcon)
     | ((v,a,b):_) <- checkDoubleVars = Just $ a ++ " " ++ v ++ " occurs also as " ++ b 
     | ((v,var,eq):_) <- checkEquations = Just $ var ++ " " ++ v ++ " occurs on both sides of the equation " ++ eq
     | ((u,v,eq):_) <- checkChains = Just $ "chain-variables on both sides of the equation " ++ eq ++ " (" ++ u ++ " on the left hand side and " ++ v ++ " on the right hand side)"
     | ((v:_):_) <- checkExprVars = Just $ "expression variable " ++ v ++ " occurs more than twice in the unification problem"
     | ((v:_,dec):_) <- checkOtherVars = Just $ dec ++ " " ++ v ++ " occurs more than once in the unification problem"
     | Just err <- checkLVCGamma gamma = Just $ "LVC not valid:" ++ err
     | otherwise = Nothing
   where  
    nonemptycxtvars = [v | MetaCtxt s srt v <- ctxtcon]
    eq1 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- varLiftEq gamma]
    eq3 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- expressionEq gamma]
    eq4 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- bindingEq gamma]
    eq5 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- environmentEq gamma]
    allVars = combineMetaVarsList $
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq1]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq3]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq4]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq5]
    checkNonEmptyCtxtVars = [v | v <- nonemptycxtvars, v `notElem` (ctxtvars allVars)]
    checkNonEmptyEnvVars = [v | Env s v <- nonemptyenvvars, v `notElem` (envvars allVars)]
    checkDoubleVars =    [(v,"context-variable","bound variable")      | v <- (ctxtvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"context-variable","expression variable")        | v <- (ctxtvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"context-variable","environment-variable")       | v <- (ctxtvars allVars), v `elem` (envvars allVars)]
                      ++ [(v,"context-variable","chain-variable")             | v <- (ctxtvars allVars), v `elem` (chainvars allVars)]
                      ++ [(v,"chain-variable","bound variable")        | v <- (chainvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"chain-variable","expression variable")          | v <- (chainvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"chain-variable","environment-variable")         | v <- (chainvars allVars), v `elem` (envvars allVars)]
                      ++ [(v,"environment-variable","bound variable")  | v <- (envvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"environment-variable","expression variable")    | v <- (envvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"expression-variable","bound variable")   | v <- (exprvars allVars), v `elem` (liftvars allVars)]

    checkEquations = concat $
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq1, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq3, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq4, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                                              ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq5, let eq = show (l :=?=: r)]                   
                         
    checkChains =     [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq1, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq3, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq4, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq5, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                                      
    checkExprVars = filter (\x -> length x > 2) $ group $ sort $ exprvars allVars
    checkOtherVars = concat [zip (filter (\x -> length x > 1) $ group $ sort $ vars) (repeat dec)
                         | (vars,dec) <- [(envvars allVars,"environment-variable"),(ctxtvars allVars,"context-variable"),(chainvars allVars,"chain-variable")]]
-- ----------------------------------
-- several helper functions on Gamma
-- ----------------------------------
environmentVarsInGamma :: Gamma -> [EnvVar]
environmentVarsInGamma gam = 
  concat (
      [ search e1 ++ search e2 | e1 :=?=: e2 <- expressionEq gam]
   ++ [ searchEnv env1 ++ searchEnv env2 | env1 :=?=: env2 <- environmentEq gam]
   ++ [ searchBind b1 ++ searchBind b2 | (b1 :=?=: b2) <- bindingEq gam]
  )
 where
    searchArgPos (Binder v e) = search e
    searchArgPos (VarPos _) = []
    search (Fn _ args) = concatMap searchArgPos args
    search (Letrec env e) = searchEnv env ++ search e
    search (FlatLetrec env e) = searchEnv env ++ search e
    search (MetaCtxtSubst _ _ _ e) = search e
    search (ConstantCtxtSubst _ _ _ e) = search e    
    search _ = []
    searchEnv env =
      let bs = concatMap searchBind $ bindings env
          ms = [Env sub v | Env sub v <- metaEnvVars env] 
          ms' = [Env sub v | Env sub v <- constantEnvVars env] 
          me = concat [search e | MetaChain _ _ _ _ e <- metaChains env]
          ce = concat [search e | ConstantChain _ _ _ _ e <- constantChains env]          
      in bs ++ ms ++ ms' ++ me ++ ce

    searchBind (_ := e) = search e

envVarsInSol  sol = concatMap getEnvvars sol
 where
     getEnvvars (MapVarLift _) = []
     getEnvvars (MapVarLet _) = []
     getEnvvars (MapExpr    (e1 :|-> e2 )) = envVarsInExpr e2
     getEnvvars  (MapEnv     (envVar :|-> env)) = envVarsInEnv env
     getEnvvars (MapCV _) = []
     getEnvvars (MapEnvSpec     (envVar :|-> env)) = envVarsInEnv env
    
envVarsInArgPos (Binder v e) = envVarsInExpr e
envVarsInArgPos (VarPos _) = []
envVarsInExpr :: Expression -> [EnvVar]
envVarsInExpr (Fn _ args) = concatMap envVarsInArgPos args
envVarsInExpr (Letrec env e) = envVarsInEnv env ++ envVarsInExpr e
envVarsInExpr (FlatLetrec env e) = envVarsInEnv env ++ envVarsInExpr e
envVarsInExpr (MetaCtxtSubst sub srt name expr) = (envVarsInExpr expr)
envVarsInExpr (ConstantCtxtSubst sub srt name expr) = (envVarsInExpr expr)
envVarsInExpr (MetaCtxt sub srt name) = []
envVarsInExpr (ConstantCtxt sub srt name) = []
envVarsInExpr _ = []
envVarsInEnv :: Environment -> [EnvVar]
envVarsInEnv env =
      let bs = concatMap envVarsInBind $ bindings env
          me = concat [envVarsInExpr e | MetaChain _ _ _ _ e <- metaChains env]
          ce = concat [envVarsInExpr e | ConstantChain _ _ _ _ e <- constantChains env]          
      in bs ++ me ++ ce ++ (metaEnvVars env) ++ (constantEnvVars env)
envVarsInBind (_ := e) = envVarsInExpr e    


    
contextVarsInSol  sol = concatMap getContextvars sol
 where
     getContextvars (MapVarLift _) = []
     getContextvars (MapVarLet _) = []
     getContextvars (MapExpr    (e1 :|-> e2 )) = contextVarsInExpr e2
     getContextvars (MapEnv     (envVar :|-> env)) = contextVarsInEnv env
     getContextvars  (MapCV _) = []
     getContextvars (MapEnvSpec     (envVar :|-> env)) = contextVarsInEnv env




contextVarsInGamma :: Gamma -> [Context]
contextVarsInGamma gam = 
  concat (
      [ contextVarsInExpr e1 ++ contextVarsInExpr e2 | e1 :=?=: e2 <- expressionEq gam]
   ++ [ contextVarsInEnv env1 ++ contextVarsInEnv env2 | env1 :=?=: env2 <- environmentEq gam]
   ++ [ contextVarsInBind b1 ++ contextVarsInBind b2 | (b1 :=?=: b2) <- bindingEq gam]
  )
 where
     
     
contextVarsInArgPos (Binder v e) = contextVarsInExpr e
contextVarsInArgPos (VarPos _) = []
contextVarsInExpr (Fn _ args) = concatMap contextVarsInArgPos args
contextVarsInExpr (Letrec env e) = contextVarsInEnv env ++ contextVarsInExpr e
contextVarsInExpr (FlatLetrec env e) = contextVarsInEnv env ++ contextVarsInExpr e
contextVarsInExpr (MetaCtxtSubst sub srt name expr) = (MetaCtxt sub srt name):(contextVarsInExpr expr)
contextVarsInExpr (ConstantCtxtSubst sub srt name expr) = (ConstantCtxt sub srt name):(contextVarsInExpr expr)
contextVarsInExpr (MetaCtxt sub srt name) = [(MetaCtxt sub srt name)]
contextVarsInExpr (ConstantCtxt sub srt name) = [(ConstantCtxt sub srt name)]
contextVarsInExpr _ = []
contextVarsInEnv env =
      let bs = concatMap contextVarsInBind $ bindings env
          me = concat [contextVarsInExpr e | MetaChain _ _ _ _ e <- metaChains env]
          ce = concat [contextVarsInExpr e | ConstantChain _ _ _ _ e <- constantChains env]          
      in bs ++ me ++ ce
contextVarsInBind (_ := e) = contextVarsInExpr e    
    
    
    
   
   
-- check whether Gamma fulfills the weak distinct variable convention

checkLVCGamma :: Gamma -> Maybe String
checkLVCGamma gamma
 | any isJust l1 = case head (filter isJust l1)
                     of Just s -> Just (s ++ " (case 1)")
 | any isJust l2 = case head (filter isJust l2)
                     of Just s -> Just (s ++ " (case 2)")
 | any isJust l3 = case head (filter isJust l3)
                     of Just s -> Just (s ++ " (case 3)")
 | otherwise     = Nothing
 where
   l1 =  concatMap (\(x :=?=: y) -> [checkLVC x, checkLVC y]) (expressionEq gamma)
   l2 =  concatMap (\(x :=?=: y) -> [checkLVCEnv x, checkLVCEnv y]) (environmentEq gamma)
   l3 =  concatMap (\(x :=?=: y) -> [checkLVCEnv empty{bindings=[x]}, checkLVCEnv empty{bindings=[y]}]) (bindingEq gamma)
 
-- ============================================================================

-- ============================================================================
-- Delta: Several constraints on the problem 
-- ============================================================================
data Delta = Delta {delta1  :: SetOfCtxtVars
                   ,delta2  :: SetOfEnvironmentVars
                   ,delta3  :: NonCaptureConstraints
                   ,delta4  :: Gamma 
                   }
 deriving(Show)
 
emptyDelta = Delta {delta1 = [], delta2 =[], delta3=[],delta4=emptyGamma} 


instance HasMetaVars Delta where
  computeMetaVars del = 
                      combineMetaVarsList [computeMetaVarsDelta1 (delta1 del)
                                          ,computeMetaVarsDelta2 (delta2 del)
                                          ,computeMetaVarsDelta3 (delta3 del)
                                          ,computeMetaVars (delta4 del)]
computeMetaVarsDelta3 l = combineMetaVarsList (map computeMetaVarsNCC l)
computeMetaVarsNCC (Left e,ctxt) = (computeMetaVars e) `combineMetaVars` (computeMetaVars ctxt)
computeMetaVarsNCC (Right e,ctxt) = (computeMetaVars e) `combineMetaVars` (computeMetaVars ctxt)
computeMetaVarsDelta2 del2 = emptyMVars {envvars = [e | Env sub e <- del2]}
computeMetaVarsDelta1 del1 = combineMetaVarsList (map computeMetaVars del1)


variablesInDelta3 list = concat $
 [ case l of 
        Left e -> allVariables e ++ allVariables r
        Right env -> allVariablesInEnv env ++ allVariables  r
  | (l,r) <- list] 

  
-- Substitute-instance
instance Substitute Delta where
 subst e1 e2 delta =
  delta { delta1 = delta1 delta
        , delta2 = delta2 delta
        , delta3 = subst e1 e2 (delta3 delta)
        , delta4 = subst e1 e1 (delta4 delta)
        }
 substEmpty n  delta =
  delta { delta1 = delta1 delta
        , delta2 = delta2 delta
        , delta3 = substEmpty n (delta3 delta)
        , delta4 = substEmpty n (delta4 delta)
        }
        
 substEnv env n  delta =
  delta { delta1 = delta1 delta
        , delta2 = delta2 delta
        , delta3 = substEnv env n (delta3 delta)
        , delta4 = substEnv env n (delta4 delta)
        }
 substChainVar env n  delta =
  delta { delta1 = delta1 delta
        , delta2 = delta2 delta
        , delta3 = substChainVar env n (delta3 delta)
        , delta4 = substChainVar env n (delta4 delta)
        }
 
 substEnvConstraint env n  delta =
  delta { delta1 = delta1 delta
        , delta2 = delta2 delta
        , delta3 = delta3 delta -- substChainVar env n (delta3 delta)
        , delta4 = delta4 delta -- substChainVar env n (delta4 delta)
        }
        
-- ============================================================================
instance ApplySolution Delta where
     applySolution sol delta =
      delta{
            delta3  = applySolution  sol (delta3 delta)
           ,delta4  = applySolution  sol (delta4 delta)
           }     
instance ApplySolution Gamma where
    applySolution sol gamma = 
     gamma{varLiftEq = applySolution sol (varLiftEq gamma)
          ,expressionEq = applySolution sol (expressionEq gamma)
          ,bindingEq = applySolution sol (bindingEq gamma)
          }
    
instance HasMetaVars Gamma where
    computeMetaVars gamma = 
         combineMetaVarsList [computeMetaVars (varLiftEq gamma)
                             ,computeMetaVars (expressionEq gamma)
                             ,computeMetaVars (bindingEq gamma)
                             ]

standardSubst (f:fresh) nabla v@(SCVar (ConstantVarLift sub name)) = 
   (MapVarLift $ (ConstantVarLift sub name) :|-> (VarLift emptySubstitution ("x_" ++ f)), fresh)
   
standardSubst (f:fresh) nabla v@(SCExpr (Constant sub name)) = 
  let v = VarLift emptySubstitution ("x_" ++ f)
  in (MapExpr $ (Constant sub name) :|-> Fn "\\" [Binder [v] (VarL v)], fresh)

standardSubst (fresh) nabla v@(SCCtxt (ConstantCtxt sub typ name)) 
  | (ConstantCtxt sub typ name) `elem` (delta1 nabla)
    ||
    (ConstantCtxt emptySubstitution typ name) `elem` (delta1 nabla)
   = (MapExpr $ (ConstantCtxt sub typ name) :|-> (Fn "__internal_dummy_f" [Binder [] (Hole emptySubstitution)]), fresh)
  | otherwise = ((MapExpr $ (ConstantCtxt sub typ name) :|-> Hole emptySubstitution), fresh)
  
standardSubst (fresh) nabla v@(SCConstantChain sub typ name) =
   ((MapCV $ (ConstantChain sub typ name undefined undefined) :|-> \z s -> empty{bindings=[z := s]}),fresh) 
   
standardSubst (f:fresh) nabla v@(SCConstantEnv (Env sub e))
  |   (Env sub e) `elem` (delta2 nabla)
   || (Env emptySubstitution e) `elem` (delta2 nabla)
  = 
   let v = VarLift emptySubstitution ("x_" ++ f)
   in ((MapEnv $ (Env sub e) :|-> (empty{bindings =  [v := (VarL v)]})),fresh) 
  | otherwise = (MapEnv ((Env sub e) :|-> empty), (f:fresh))

mkStandardSubst fresh nabla [] = ([],fresh)
mkStandardSubst fresh nabla (v:vs) =
  let (sub,fresh')   = standardSubst fresh nabla v 
      (subs,fresh'') = mkStandardSubst fresh' nabla vs 
  in (sub:subs,fresh'')
  
instance ApplySolution a => ApplySolution (Equation a) where
    applySolution s (a :=?=: b) = (applySolution s a) :=?=: (applySolution s b)
  